"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const StreamArchive_1 = require("./StreamArchive");
const DMArchive_1 = __importDefault(require("./DMArchive"));
const event_target_shim_1 = require("event-target-shim");
const big_integer_1 = __importDefault(require("big-integer"));
const helpers_1 = require("./helpers");
const jszip_1 = __importDefault(require("jszip"));
const js_md5_1 = __importDefault(require("js-md5"));
/** Return the `Date` object affiliated to **tweet**. */
function dateFromTweet(tweet) {
    if (tweet.created_at_d && tweet.created_at_d instanceof Date) {
        return tweet.created_at_d;
    }
    return tweet.created_at_d = parseTwitterDate(tweet.created_at);
}
exports.dateFromTweet = dateFromTweet;
function parseTwitterDate(date) {
    try {
        const d = new Date(date);
        if (isNaN(d.getTime())) {
            throw "";
        }
        else {
            return d;
        }
    }
    catch (e) {
        return new Date(date.replace(/\s(.+)\s.+/, 'T$1.000Z'));
    }
}
exports.parseTwitterDate = parseTwitterDate;
/**
 * Return true if **tweet** contains media(s).
 *
 * This includes photos, videos or animated GIF.
 */
function isWithMedia(tweet) {
    return tweet.entities.media.length > 0;
}
exports.isWithMedia = isWithMedia;
/**
 * Return true if **tweet** contains a video or one animated GIF.
 *
 * Twitter's GIF are mp4 encoded.
 */
function isWithVideo(tweet) {
    if (tweet.extended_entities) {
        if (tweet.extended_entities.media) {
            return tweet.extended_entities.media.some(m => m.type !== "photo");
        }
    }
    return false;
}
exports.isWithVideo = isWithVideo;
/**
 * Represents a full TwitterArchive. Support GDPR and classic archive.
 *
 * Remember that tweets, in searchs in particular are **NOT** sorted.
 *
 * Direct messages, parsed if archive is a GDPR archive, stored in `.messages`,
 * are returned and sorted from the most older to the more recent.
 */
class TwitterArchive extends event_target_shim_1.EventTarget {
    /**
     * Build a new instance of TwitterArchive.
     *
     * @param file Accept any source that `JSZip` library accepts, and string for filenames.
     * If this parameter is `null`, then you'll need to load each part one by one !
     * @param build_extended True if `.extended_gdpr` should be built (only if **file** is a GDPR archive.)
     * @param keep_loaded If possible, free the memory after load if set to false.
     * @param load_images_in_zip In Twitter GDPR archives v2, tweet and dm images are in ZIP archives inside the ZIP.
     * If `true`, TwitterArchive will extract its content in RAM to allow the usage of images.
     * If `false`, DMs images will be unavailable.
     * If `undefined`, Twitter will extract in RAM in browser mode, and leave the ZIP untouched in Node.js.
     *
     * If you want to save memory, set this parameter to `false`,
     * and before using `.dmImage()` methods, check if you need to load DM images ZIP
     * with `.requiresDmImageZipLoad()`.
     *
     * Then, if you need to, load the DM image ZIP present in the archive using `.loadCurrentDmImageZip()`.
     * **Please note that `keep_loaded` should be set to `true` to use this method !**
     */
    constructor(file, build_extended = true, keep_loaded = false, load_images_in_zip = undefined) {
        super();
        this.load_images_in_zip = load_images_in_zip;
        this._ready = Promise.resolve();
        /** Current archive load state. */
        this.state = "idle";
        this._index = {
            info: {
                screen_name: "",
                full_name: "",
                location: "fr",
                bio: "",
                id: "",
                created_at: (new Date).toISOString()
            },
            archive: {
                created_at: (new Date).toISOString(),
                tweets: 0
            },
            years: {},
            by_id: {}
        };
        this._is_gdpr = false;
        if (file === null) {
            this.state = "ready";
        }
        else {
            this.state = "reading";
            this._ready =
                (file instanceof Promise ?
                    file.then(f => (this.archive = StreamArchive_1.constructArchive(f)).ready())
                    : (this.archive = StreamArchive_1.constructArchive(file)).ready())
                    .then(() => {
                    this.dispatchEvent({ type: 'zipready' });
                    // Initialisation de l'archive Twitter
                    if (this.isGDPRArchive()) {
                        return this.initGDPR(build_extended, keep_loaded);
                    }
                    else {
                        return this.initClassic().then(() => {
                            if (!keep_loaded) {
                                this.archive = undefined;
                            }
                        });
                    }
                })
                    .then(() => {
                    this.dispatchEvent({ type: 'ready' });
                })
                    .catch(e => {
                    this.dispatchEvent({ type: 'error', detail: e });
                    return Promise.reject(e);
                });
        }
    }
    async initGDPR(extended, keep_loaded) {
        try {
            // Delete the tweet media folder (big & useless)
            if (this.archive instanceof StreamArchive_1.Archive) {
                this.archive.raw.remove('tweet_media');
            }
        }
        catch (e) { }
        // This is not accurate, but this is for compatibility reasons
        this.dispatchEvent({ type: 'userinfosready' });
        this.state = "tweet_read";
        // ------------------------
        // TWEETS AND PROFILE INFOS
        // ------------------------
        // Init tweet indexes
        const tweets = await this.archive.get('tweet.js');
        let i = 1;
        while (this.archive.has(`tweet-part${i}.js`)) {
            // Add every tweet in other files 
            // inside a "new" array in the initial array
            tweets.push(...await this.archive.get(`tweet-part${i}.js`));
            i++;
        }
        this.dispatchEvent({ type: 'tweetsread' });
        this.loadArchivePart({
            account: await this.archive.get('account.js'),
            profile: await this.archive.get('profile.js'),
            tweets
        });
        this.dispatchEvent({ type: 'indexready' });
        // ---------------
        // DIRECT MESSAGES
        // ---------------
        this.state = "dm_read";
        this.dispatchEvent({ type: 'willreaddm' });
        // Init DMs
        const conv_files = ['direct-message.js', 'direct-message-group.js'];
        i = 1;
        while (this.archive.has(`direct-message-part${i}.js`)) {
            conv_files.push(`direct-message-part${i}.js`);
            i++;
        }
        this.loadArchivePart({
            dms: await Promise.all(conv_files
                .filter(name => this.archive.has(name))
                .map(file => this.archive.get(file))),
        });
        // DMs should be ok
        // Test if archive contain DM images as zip
        let can_unload_archive = !keep_loaded;
        if (this.archive.searchDir(/direct_message_media/).length) {
            const folder = this.archive.dir('direct_message_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                // console.log("Creating archive from archive (single)");
                if (this.should_autoload_zip_img)
                    this.dm_img_archive = await folder.fromFile(query[0]);
            }
            else {
                // cannot unload
                can_unload_archive = false;
            }
        }
        if (this.archive.searchDir(/direct_message_group_media/).length) {
            const folder = this.archive.dir('direct_message_group_media');
            const query = folder.search(/\.zip$/);
            if (query.length && this.should_autoload_zip_img) {
                // console.log("Creating archive from archive (group)");
                this.dm_img_group_archive = await folder.fromFile(query[0]);
            }
        }
        this.state = "extended_read";
        this.dispatchEvent({ type: 'willreadextended' });
        if (extended) {
            await this.initExtendedGDPR();
        }
        this.state = "ready";
        if (can_unload_archive) {
            this.archive = undefined;
        }
    }
    async initExtendedGDPR() {
        // Followings
        const followings = new Set();
        try {
            const f_following = await this.archive.get('following.js');
            for (const f of f_following) {
                followings.add(f.following.accountId);
            }
        }
        catch (e) { }
        // Followers
        const followers = new Set();
        try {
            const f_follower = await this.archive.get('follower.js');
            for (const f of f_follower) {
                followers.add(f.follower.accountId);
            }
        }
        catch (e) { }
        // Favorites
        const favorites = new Set();
        try {
            const f_fav = await this.archive.get('like.js');
            for (const f of f_fav) {
                favorites.add(f.like.tweetId);
            }
        }
        catch (e) { }
        // Mutes
        const mutes = new Set();
        try {
            const f_mutes = await this.archive.get('mute.js');
            for (const f of f_mutes) {
                mutes.add(f.muting.accountId);
            }
        }
        catch (e) { }
        // Blocks
        const blocks = new Set();
        try {
            const f_block = await this.archive.get('block.js');
            for (const f of f_block) {
                blocks.add(f.blocking.accountId);
            }
        }
        catch (e) { }
        // Lists
        const lists = {
            created: [],
            member_of: [],
            subscribed: []
        };
        try {
            lists.created = (await this.archive.get('lists-created.js'))[0].userListInfo.urls;
            lists.member_of = (await this.archive.get('lists-member.js'))[0].userListInfo.urls;
            lists.subscribed = (await this.archive.get('lists-subscribed.js'))[0].userListInfo.urls;
        }
        catch (e) { }
        // Personalization
        let personalization = {
            demographics: undefined,
            locationHistory: [],
            interests: undefined
        };
        try {
            personalization = (await this.archive.get('personalization.js'))[0].p13nData;
        }
        catch (e) { }
        let age_info;
        try {
            age_info = (await this.archive.get('ageinfo.js'))[0].ageMeta;
        }
        catch (e) { }
        // SN history
        const screen_name_history = [];
        try {
            const f_history = await this.archive.get('screen-name-change.js');
            for (const e of f_history) {
                screen_name_history.push(e.screenNameChange);
            }
        }
        catch (e) { }
        // Protected history
        const protected_history = [];
        try {
            const f_phistory = await this.archive.get('protected-history.js');
            for (const e of f_phistory) {
                protected_history.push(e.protectedHistory);
            }
        }
        catch (e) { }
        // Moments
        const moments = (await this.archive.get('moment.js')).map(e => e.moment);
        this.extended_gdpr = {
            moments,
            protected_history,
            screen_name_history,
            age_info,
            personalization,
            lists,
            favorites,
            followers,
            followings,
            mutes,
            blocks
        };
    }
    convertToPartial(tweet, pic_prof) {
        if (!this.user_cache) {
            this.user_cache = {
                protected: false,
                id_str: this._index.info.id,
                name: this._index.info.full_name,
                screen_name: this._index.info.screen_name,
                profile_image_url_https: pic_prof
            };
        }
        tweet.user = this.user_cache;
        tweet.text = tweet.full_text;
        // @ts-ignore
        tweet.retweet_count = Number(tweet.retweet_count);
        // @ts-ignore
        tweet.favorite_count = Number(tweet.favorite_count);
        // Gérer le cas des retweets
        const rt_data = /^RT @(.+?): (.+)/.exec(tweet.full_text);
        if (rt_data && rt_data.length && !("retweeted_status" in tweet)) {
            const [, arobase, text] = rt_data;
            const rt = Object.assign({}, tweet);
            rt.user = Object.assign({}, rt.user);
            rt.text = text;
            rt.user.screen_name = arobase;
            rt.user.name = arobase;
            // @ts-ignore
            rt.retweeted = true;
            rt.retweet_count = Number(rt.retweet_count);
            rt.favorite_count = Number(rt.favorite_count);
            // Recherche si un ID est disponible par exemple dans les medias (sinon tant pis)
            if (rt.extended_entities && rt.extended_entities.media) {
                if (rt.extended_entities.media.length) {
                    rt.user.id_str = rt.extended_entities.media[0].source_user_id_str;
                }
            }
            tweet.retweeted_status = rt;
        }
        return tweet;
    }
    async initClassic() {
        this.state = "user_read";
        const js_dir = this.archive.dir('data').dir('js');
        const index = await js_dir.get('tweet_index.js');
        this.loadClassicArchivePart({
            payload: await js_dir.get('payload_details.js'),
            user: await js_dir.get('user_details.js'),
        });
        this.dispatchEvent({ type: 'userinfosready' });
        const files_to_read = index.map(e => e.file_name);
        this.state = "tweet_read";
        const tweet_file_promises = [];
        for (const file of files_to_read) {
            tweet_file_promises.push(this.archive.get(file));
        }
        let tweets = [].concat(...await Promise.all(tweet_file_promises));
        // Tri les tweets par ID (le plus récent, plus grand en premier)
        tweets = this.sortTweets(tweets);
        this.dispatchEvent({ type: 'tweetsread' });
        this.state = "indexing";
        // Build index (read tweets)
        this.readTweets(tweets);
        this.dispatchEvent({ type: 'indexready' });
        this.state = "ready";
    }
    isGDPRArchive() {
        return this._is_gdpr = this.archive.search(/^tweets\.csv$/).length === 0;
    }
    /**
     * True if given tweet is coming from a GDPR archive.
     *
     * Tweets getted by available getters are NOT GDPR tweets, they've been converted !
     */
    isGDPRTweet(tweet) {
        return 'retweet_count' in tweet;
    }
    /** Extract tweets from a specific month. */
    month(month, year) {
        if (year in this.index.years) {
            if (month in this.index.years[year]) {
                return Object.values(this.index.years[year][month]);
            }
        }
        return [];
    }
    /** Find tweets made on the same day (= month, = day), but in all years. */
    fromThatDay() {
        const now = new Date;
        const now_m = now.getMonth();
        const now_d = now.getDate();
        // TODO optimize
        return this.all.filter(t => {
            const d = dateFromTweet(t);
            return d.getMonth() === now_m && d.getDate() === now_d;
        });
    }
    /** Get tweets in a specific time interval. */
    between(since, until) {
        if (since.getTime() > until.getTime()) {
            throw new Error("Since can't be superior to until");
        }
        const firsts = [String(since.getMonth() + 1), String(since.getFullYear())];
        const ends = [String(until.getMonth() + 1), String(until.getFullYear())];
        const remaining_months = [];
        remaining_months.push(firsts);
        let tmp_date = new Date(since.getFullYear(), since.getMonth());
        tmp_date.setMonth(tmp_date.getMonth() + 1);
        remaining_months.push(ends);
        // Calculating months that are between the two dates
        while (true) {
            if (tmp_date.getFullYear() > until.getFullYear()) {
                break;
            }
            if (tmp_date.getFullYear() === until.getFullYear()) {
                if (tmp_date.getMonth() >= until.getMonth()) {
                    break;
                }
            }
            remaining_months.push([String(tmp_date.getMonth() + 1), String(tmp_date.getFullYear())]);
            tmp_date.setMonth(tmp_date.getMonth() + 1);
        }
        const tweets = [];
        const [end_day, end_month, end_year] = [until.getDate(), until.getMonth(), until.getFullYear()];
        // Finding tweets that are same or after since date or before until date
        for (const [month, year] of remaining_months) {
            const m = this.month(month, year);
            for (const t of m) {
                const d = dateFromTweet(t);
                if (d.getTime() >= since.getTime()) {
                    if (d.getFullYear() < end_year || d.getMonth() < end_month || d.getDate() <= end_day) {
                        tweets.push(t);
                    }
                }
            }
        }
        return tweets;
    }
    /** Get a single tweet by ID. Returns `null` if tweet does not exists. */
    id(id_str) {
        if (id_str in this.index.by_id) {
            return this.index.by_id[id_str];
        }
        return null;
    }
    /**
     * Give the media url in direct message, obtain the Blob-bed image.
     * For use in Node.js, you must set `as_array_buffer` to `true` !
     */
    dmImageFromUrl(url, is_group = false, as_array_buffer = false) {
        const [, , , , id, , image] = url.split('/');
        if (id && image) {
            return this.dmImage(id + "-" + image, is_group, as_array_buffer);
        }
        return Promise.reject("URL is invalid");
    }
    /**
     * Extract a direct message image from GDPR archive (exact filename required).
     * For use in Node.js, you must set `as_array_buffer` to `true` !
     */
    dmImage(name, is_group = false, as_array_buffer = false) {
        if (!this.is_gdpr) {
            return Promise.reject("Archive not supported");
        }
        if (this.dm_img_archive || this.dm_img_group_archive) {
            // Les dm sont dans un ZIP
            let zip;
            if (is_group) {
                zip = this.dm_img_group_archive;
            }
            else {
                zip = this.dm_img_archive;
            }
            if (!zip) {
                return Promise.reject("Image archive is not loaded, impossible to get media.");
            }
            const results = zip.search(new RegExp(name + "(\.?.*)$"));
            if (results.length) {
                return zip.read(results[0], as_array_buffer ? "arraybuffer" : "blob");
            }
            return Promise.reject("File not found");
        }
        else {
            if (!this.archive) {
                return Promise.reject("Archive is not available/loaded: DM images are not fetchables");
            }
            const directory = this.archive.dir(is_group ? "direct_message_group_media" : "direct_message_media");
            const results = directory.search(new RegExp(name + "(\.?.*)$"));
            if (results.length) {
                return this.archive.read(results[0], as_array_buffer ? "arraybuffer" : "blob");
            }
            return Promise.reject("File not found");
        }
    }
    /**
     * Return all the images of a direct message, as blob or array buffer.
     *
     * If the message does not exists or the DM archive is not loaded / available,
     * return an empty array.
     *
     * Otherwise, return a array of `Blob` / `ArrayBuffer`
     *
     * @param direct_message Direct Message id | Direct message
     * @param as_array_buffer Return an `ArrayBuffer` array, instead of a `Blob` array
     */
    async dmImagesOf(direct_message, as_array_buffer = false) {
        if (!this.is_gdpr || !this.messages) {
            return [];
        }
        if (typeof direct_message === 'string') {
            direct_message = this.messages.single(direct_message);
        }
        if (!direct_message) {
            // Message not found
            return [];
        }
        const images = [];
        for (const media of direct_message.mediaUrls) {
            images.push(this.dmImageFromUrl(media, as_array_buffer));
        }
        return Promise.all(images);
    }
    /**
     * Return true if you need to load a DM image ZIP in order to use `.dmImage()`.
     *
     * If you need to, see `.loadCurrentDmImageZip()`, or `importDmImageZip()`.
     */
    requiresDmImageZipLoad() {
        if (!this.is_gdpr || !this.archive || this.dm_img_archive) {
            return false;
        }
        if (this.archive.searchDir(new RegExp('direct_message_media')).length) {
            const folder = this.archive.dir('direct_message_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                return true;
            }
        }
        return false;
    }
    /**
     * Load/reload zip archives that contains DM images,
     * if `load_images_in_zip` parameter was set to false.
     *
     * You need to have the archive loaded to accomplish this action
     * (constructor `keep_loaded` parameter should be set to `true`).
     *
     * Note that if any zip is found in the archive, this method will just do nothing.
     */
    async loadCurrentDmImageZip() {
        if (!this.archive || !this.is_gdpr) {
            return;
        }
        if (this.archive.searchDir(new RegExp('direct_message_media')).length) {
            const folder = this.archive.dir('direct_message_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                // Load the archive
                this.dm_img_archive = await folder.fromFile(query[0]);
                await this.dm_img_archive.ready();
            }
        }
        if (this.archive.searchDir(new RegExp('direct_message_group_media')).length) {
            const folder = this.archive.dir('direct_message_group_media');
            const query = folder.search(/\.zip$/);
            if (query.length) {
                // Load the archive (group)
                this.dm_img_group_archive = await folder.fromFile(query[0]);
                await this.dm_img_group_archive.ready();
            }
        }
    }
    /**
     * Import a custom ZIP file as DM single-conversation images file.
     */
    async importDmImageZip(file) {
        this.dm_img_archive = StreamArchive_1.constructArchive(await file);
        await this.dm_img_archive.ready();
    }
    /**
     * Import a custom ZIP file as DM group-conversation images file.
     */
    async importDmGroupImageZip(file) {
        this.dm_img_group_archive = StreamArchive_1.constructArchive(await file);
        await this.dm_img_group_archive.ready();
    }
    /** All tweets registered in this archive. */
    get all() {
        return Object.values(this.index.by_id);
    }
    /** Access to the DMArchive object. Will be undefined if `.is_gdpr === false`. */
    get messages() {
        return this.dms;
    }
    /** ID of the user who created this archive. */
    get owner() {
        return this._index.info.id;
    }
    /**
     * Screen name (@) of the user who created this archive.
     * May be obsolete (user can change screen_name over time).
     */
    get owner_screen_name() {
        return this._index.info.screen_name;
    }
    /** Archive creation date. Not accurate in GDPR archive (will be the current date). */
    get generation_date() {
        return parseTwitterDate(this._index.archive.created_at);
    }
    /** Archive information and tweet index. */
    get index() {
        return this._index;
    }
    /** Number of tweets in this archive. */
    get length() {
        return this.index.archive.tweets;
    }
    /** True if archive is a GDPR archive. */
    get is_gdpr() {
        if (this._is_gdpr === undefined)
            return this.isGDPRArchive();
        return this._is_gdpr;
    }
    /** Raw Archive object. Can be used to get specific files. */
    get raw() {
        return this.archive;
    }
    get should_autoload_zip_img() {
        if (typeof this.load_images_in_zip !== 'undefined') {
            return this.load_images_in_zip;
        }
        // test if node
        if (typeof process !== 'undefined' && process.versions && process.versions.node) {
            // IsNode
            return false;
        }
        return true;
    }
    /** Resolved when archive read is over. */
    ready() {
        return this._ready;
    }
    /** -------------------- */
    /** SIDELOADING MANUALLY */
    /** -------------------- */
    /**
     * Load a part of a GDPR archive.
     *
     * Set current archive as GDPR archive.
     */
    loadArchivePart(parts = {}) {
        this._is_gdpr = true;
        if (parts.account) {
            // Init informations
            const account = parts.account[0].account;
            this._index.info.screen_name = account.username;
            this._index.info.full_name = account.accountDisplayName;
            this._index.info.id = account.accountId;
            this._index.info.created_at = account.createdAt;
        }
        if (parts.profile) {
            const profile = parts.profile[0].profile;
            this._index.info.location = profile.description.location;
            this._index.info.bio = profile.description.bio;
            this._index.info.profile_image_url_https = profile.avatarMediaUrl;
        }
        if (parts.tweets) {
            this.readGDPRTweets(parts.tweets, this.index.info.profile_image_url_https);
        }
        if (parts.dms) {
            if (!this.dms) {
                this.dms = new DMArchive_1.default(this.owner);
            }
            for (const file of parts.dms) {
                this.dms.add(file);
            }
        }
    }
    /**
     * Load a part of a classic archive.
     *
     * Set current archive as non-gdpr archive.
     */
    loadClassicArchivePart(parts = {}) {
        this._is_gdpr = false;
        if (parts.tweets) {
            this.readTweets(this.sortTweets(parts.tweets));
        }
        if (parts.user) {
            this.index.info = parts.user;
        }
        if (parts.payload) {
            this.index.archive = parts.payload;
        }
    }
    /**
     * Convert GDPR tweets into classic tweets, then register them into the index.
     */
    readGDPRTweets(tweets, profile_img_https) {
        // Build index
        for (let i = 0; i < tweets.length; i++) {
            const date = parseTwitterDate(tweets[i].created_at);
            const month = String(date.getMonth() + 1);
            const year = String(date.getFullYear());
            // Creating month/year if not presents
            if (!(year in this.index.years)) {
                this.index.years[year] = {};
            }
            if (!(month in this.index.years[year])) {
                this.index.years[year][month] = {};
            }
            // Save tweet in index
            const converted = this.convertToPartial(tweets[i], profile_img_https);
            this.index.years[year][month][tweets[i].id_str] = converted;
            this.index.by_id[tweets[i].id_str] = converted;
        }
        // Setting right tweet number
        this.index.archive.tweets = Object.keys(this.index.by_id).length;
    }
    /**
     * Register tweets into the index.
     */
    readTweets(tweets) {
        // Build index (read tweets)
        for (let i = 0; i < tweets.length; i++) {
            const date = dateFromTweet(tweets[i]);
            const month = String(date.getMonth() + 1);
            const year = String(date.getFullYear());
            // Creating month/year if not presents
            if (!(year in this.index.years)) {
                this.index.years[year] = {};
            }
            if (!(month in this.index.years[year])) {
                this.index.years[year][month] = {};
            }
            // Save tweet in index
            this.index.years[year][month][tweets[i].id_str] = tweets[i];
            this.index.by_id[tweets[i].id_str] = tweets[i];
        }
        // Setting right tweet number
        this.index.archive.tweets = Object.keys(this.index.by_id).length;
    }
    /**
     * Sort tweets by ID.
     */
    sortTweets(tweets) {
        if (helpers_1.supportsBigInt()) {
            return tweets.sort((a, b) => Number(BigInt(b.id_str) - BigInt(a.id_str)));
        }
        else {
            return tweets.sort((a, b) => (big_integer_1.default(b.id_str).minus(big_integer_1.default(a.id_str))).toJSNumber());
        }
    }
    /** --------------------- */
    /** ARCHIVE EXPORT / LOAD */
    /** --------------------- */
    async exportSave() {
        const info = this.archive_save_info;
        info.hash = this.hash(info);
        function convertConversationToGDPRConversation(conversation) {
            return {
                dmConversation: {
                    conversationId: conversation.id,
                    messages: conversation.all
                        .map(message => ({
                        messageCreate: {
                            recipientId: message.recipientId,
                            createdAt: message.createdAt,
                            mediaUrls: message.mediaUrls,
                            text: message.text,
                            senderId: message.senderId,
                            id: message.id
                        }
                    }))
                }
            };
        }
        const tweets = this.all;
        for (const tweet of tweets) {
            delete tweet.created_at_d;
        }
        const tweet_zip = await new jszip_1.default().file("tweet.json", JSON.stringify(tweets)).generateAsync({
            type: "arraybuffer",
            compression: "DEFLATE",
            compressionOptions: {
                level: 6 // Not too much, if we want a good generation time
            }
        });
        const mutes = this.extended_gdpr ? [...this.extended_gdpr.mutes] : [];
        const blocks = this.extended_gdpr ? [...this.extended_gdpr.blocks] : [];
        let dms = null;
        if (this.is_gdpr && this.messages) {
            // Swallow copy all the dms, save them to a JSZip instance
            /*
              dm.json => [
                GDPRConversation,
                ...
              ]
            */
            dms = await new jszip_1.default()
                .file("dm.json", JSON.stringify(this.messages.all.map(convertConversationToGDPRConversation)))
                .generateAsync({
                type: "arraybuffer",
                compression: "DEFLATE",
                compressionOptions: {
                    level: 6 // Not too much, if we want a good generation time
                }
            });
        }
        return {
            tweets: tweet_zip,
            dms,
            info,
            mutes,
            blocks,
            screen_name_history: this.extended_gdpr ? this.extended_gdpr.screen_name_history : []
        };
    }
    get archive_save_info() {
        const info = {
            index: Object.assign({}, this._index),
            is_gdpr: this.is_gdpr,
            version: "1.0.0",
            last_tweet_date: "",
            hash: "",
            tweet_count: this.length,
            dm_count: this.messages ? this.messages.count : 0,
        };
        // Remove the real archive index
        // @ts-ignore
        delete info.index.years;
        // @ts-ignore
        delete info.index.by_id;
        // Take the last available year
        const last_year = Object.keys(this._index.years).sort((a, b) => Number(b) - Number(a))[0];
        const last_month = Object.keys(this._index.years[last_year]).sort((a, b) => Number(b) - Number(a))[0];
        const tweets = this._index.years[last_year][last_month];
        let last_date = 0;
        for (const tweet of Object.values(tweets)) {
            const cur_date = dateFromTweet(tweet).getTime();
            if (cur_date > last_date) {
                last_date = cur_date;
            }
        }
        info.last_tweet_date = new Date(last_date ? last_date : Date.now()).toString();
        return info;
    }
    hash(from) {
        const info = from ? from : this.archive_save_info;
        return js_md5_1.default(JSON.stringify({
            screen_name: info.index.info.screen_name,
            bio: info.index.info.bio,
            name: info.index.info.full_name,
            profile_image_url_https: info.index.info.profile_image_url_https,
            created_at: info.index.info.created_at,
            tweets: info.tweet_count,
            dms: info.dm_count,
            last_tweet_date: info.last_tweet_date,
            id: info.index.info.id,
            location: info.index.info.location,
        }));
    }
    static async importSave(save) {
        save = await save;
        if (!this.SUPPORTED_SAVE_VERSIONS.includes(save.info.version)) {
            throw new Error("Save version is not supported.");
        }
        const archive = new TwitterArchive(null);
        archive._index.archive = save.info.index.archive;
        archive._index.info = save.info.index.info;
        const tweet_archive = await jszip_1.default.loadAsync(save.tweets);
        let current_load_object = JSON.parse(await tweet_archive.file("tweet.json").async("text"));
        archive.readTweets(current_load_object);
        current_load_object = undefined;
        archive._is_gdpr = save.info.is_gdpr;
        if (save.dms) {
            const dm_archive = await jszip_1.default.loadAsync(save.dms);
            current_load_object = JSON.parse(await dm_archive.file("dm.json").async("text"));
            archive.loadArchivePart({
                dms: [current_load_object]
            });
        }
        if (archive.is_gdpr) {
            archive.extended_gdpr = {
                followers: new Set,
                followings: new Set,
                mutes: new Set(save.mutes),
                blocks: new Set(save.blocks),
                personalization: undefined,
                favorites: new Set,
                lists: {
                    created: [],
                    member_of: [],
                    subscribed: []
                },
                screen_name_history: save.screen_name_history,
                protected_history: [],
                age_info: undefined,
                moments: []
            };
        }
        return archive;
    }
}
exports.TwitterArchive = TwitterArchive;
TwitterArchive.SUPPORTED_SAVE_VERSIONS = ["1.0.0"];
// Define onevents
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "zipready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "userinfosready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "tweetsread");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "indexready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "willreaddm");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "willreadextended");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "ready");
event_target_shim_1.defineEventAttribute(TwitterArchive.prototype, "error");
